import sys
import logging
import re
import string
from os.path import realpath as path_realpath

_log = logging.getLogger(name=__name__)

# Regexps for Hack instructions
#
# group(1) of A instruction match returns address or label
#
# group(1) of C instruction match returns destination specification
# group(2) of C instruction match returns computation specification
# group(3) of C instruction match returns jump specification
A_INSTRUCTION = re.compile("^@([0-9]+|[a-zA-Z_.$:]+[0-9a-zA-Z_.$:]*)$")
C_INSTRUCTION = re.compile("^([AMD]+=)*([AMD!\-+|&01]+)(;[JGTEQLNMP]*)*$")
L_INSTRUCTION = re.compile("^\(([a-zA-Z_.$:]+[0-9a-zA-Z_.$:]*)*\)$")
COMMENT = re.compile("//.*")


def _cleanup_instruction(instruction):
    instruction = instruction.translate({ord(whitespace): None for whitespace in string.whitespace})
    return re.sub(COMMENT, "", instruction)


def _instruction_type(instruction):
    if (match_instruction_object := re.fullmatch(A_INSTRUCTION, instruction)) is not None:
        _log.debug(f"{instruction} is A instruction")
        return {"type": "A_INSTRUCTION", "obj": match_instruction_object}
    elif (match_instruction_object := re.fullmatch(C_INSTRUCTION, instruction)) is not None:
        _log.debug(f"{instruction} is C instruction")
        return {"type": "C_INSTRUCTION", "obj": match_instruction_object}
    elif (match_instruction_object := re.fullmatch(L_INSTRUCTION, instruction)) is not None:
        _log.debug(f"{instruction} is L (label) instruction")
        return {"type": "L_INSTRUCTION", "obj": match_instruction_object}
    else:
        sys.exit(f"{instruction} is unknown")


def parse(asm_file):
    asm_file_path = path_realpath(asm_file)

    parsed_code = list()
    try:
        with open(asm_file_path, "r") as assembly_file_descriptor:
            _log.debug(f"Start to process {asm_file_path}")
            instruction = assembly_file_descriptor.readline()
            while instruction:
                instruction = _cleanup_instruction(instruction)
                if instruction:
                    # If not empty string or comment let's parse it
                    parsed_code.append(_instruction_type(instruction))

                instruction = assembly_file_descriptor.readline()

        return parsed_code
    except FileNotFoundError:
        sys.exit(f"File {asm_file_path} not found.")



import sys
import logging
from os.path import realpath as path_realpath
from os.path import splitext as path_splitext

from assembler.symbols import get_symbol_table, comp_table, dest_table, jump_table

_log = logging.getLogger(name=__name__)
HACK_FILE_EXTENSION = '.hack'


def _a_instruction(instruction, symbol_table):
    address = instruction['obj'].group(1)

    _log.debug(f"Produce @{address} code")
    try:
        address = int(address)
        return '{0:016b}'.format(address)
    except ValueError:
        address = symbol_table[address]
        address = int(address)
        return '{0:016b}'.format(address)


def _c_instruction(instruction):
    instruction_tuple = instruction['obj'].groups()
    _log.debug(f"Produce {instruction_tuple} code")

    header = "111"
    comp = instruction['obj'].group(2)
    dest = instruction['obj'].group(1)
    jump = instruction['obj'].group(3)

    if comp:
        try:
            comp_field = comp_table[comp]
        except KeyError:
            sys.exit(f"Unknown computation field: {comp}")
    else:
        sys.exit(f"Computation field must be present: {instruction_tuple}")

    if dest:
        try:
            dest_field = dest_table[dest.strip('=')]
        except KeyError:
            sys.exit(f"Unknown destination field: {dest}")
    else:
        dest_field = "000"

    if jump:
        try:
            jump_field = jump_table[jump.strip(';')]
        except KeyError:
            sys.exit(f"Unknown jump field: {dest}")
    else:
        jump_field = "000"

    return header + comp_field + dest_field + jump_field


def assemble(parsed_code, file):
    symbol_table = get_symbol_table(parsed_code)

    assembled_code = list()
    for instruction in parsed_code:
        if instruction['type'] == 'A_INSTRUCTION':
            assembled_code.append(str(_a_instruction(instruction, symbol_table)))
        elif instruction['type'] == 'C_INSTRUCTION':
            assembled_code.append(str(_c_instruction(instruction)))

    hack_file = path_realpath(file)
    hack_file = path_splitext(hack_file)[0] + HACK_FILE_EXTENSION

    _log.debug(f"Write code to {hack_file}")
    with open(hack_file, "w") as hack_file_descriptor:
        for instruction in assembled_code:
            hack_file_descriptor.write(f"{instruction}\n")

    return assembled_code



import logging

_log = logging.getLogger(name=__name__)

comp_table = {
    "0": "0101010",
    "1": "0111111",
    "-1": "0111010",
    "D": "0001100",
    "A": "0110000",
    "!D": "0001101",
    "!A": "0110001",
    "-D": "0001111",
    "-A": "0110011",
    "D+1": "0011111",
    "A+1": "0110111",
    "D-1": "0001110",
    "A-1": "0110010",
    "D+A": "0000010",
    "D-A": "0010011",
    "A-D": "0000111",
    "D&A": "0000000",
    "D|A": "0010101",
    "M": "1110000",
    "!M": "1110001",
    "-M": "1110011",
    "M+1": "1110111",
    "M-1": "1110010",
    "D+M": "1000010",
    "D-M": "1010011",
    "M-D": "1000111",
    "D&M": "1000000",
    "D|M": "1010101"
}

dest_table = {
    "M": "001",
    "D": "010",
    "MD": "011",
    "A": "100",
    "AM": "101",
    "AD": "110",
    "AMD": "111"
}

jump_table = {
    "JGT": "001",
    "JEQ": "010",
    "JGE": "011",
    "JLT": "100",
    "JNE": "101",
    "JLE": "110",
    "JMP": "111"
}

VARIABLES_START_ADDRESS = 16
default_symbols_table = {
    "SP": 0,
    "LCL": 1,
    "ARG": 2,
    "THIS": 3,
    "THAT": 4,
    "R0": 0,
    "R1": 1,
    "R2": 2,
    "R3": 3,
    "R4": 4,
    "R5": 5,
    "R6": 6,
    "R7": 7,
    "R8": 8,
    "R9": 9,
    "R10": 10,
    "R11": 11,
    "R12": 12,
    "R13": 13,
    "R14": 14,
    "R15": 15,
    "SCREEN": 16384,
    "KBD": 24576
}


def get_symbol_table(parsed_code):
    # Get list of label symbols
    label_instructions = list(filter(lambda instruction: instruction['type'] == 'L_INSTRUCTION', parsed_code))
    label_symbols = [instruction['obj'].group(1) for instruction in label_instructions]
    _log.debug(f"Labels in code: {label_symbols}")

    # Build label and variable tables
    assembled_with_symbols_code = list()
    labels_symbols_table = dict(default_symbols_table)
    variables_symbols_table = dict()
    current_address = 0
    for instruction in parsed_code:
        if instruction['type'] == 'A_INSTRUCTION':
            assembled_with_symbols_code.append(instruction)
            variable_symbol = instruction['obj'].group(1)
            # Get first occurrence of variable and assign address from RAM
            if variable_symbol not in label_symbols \
                    and not variable_symbol.isdigit() \
                    and variable_symbol not in default_symbols_table \
                    and variable_symbol not in variables_symbols_table:
                variable_address = len(variables_symbols_table) + VARIABLES_START_ADDRESS
                _log.debug(f"Set address of {variable_symbol} to {variable_address}")
                variables_symbols_table[variable_symbol] = variable_address
        elif instruction['type'] == 'C_INSTRUCTION':
            assembled_with_symbols_code.append(instruction)
        elif instruction['type'] == 'L_INSTRUCTION':
            label = instruction['obj'].group(1)
            _log.debug(f"Set address of {label} to {current_address}")
            labels_symbols_table[label] = current_address

        current_address = len(assembled_with_symbols_code)

    _log.debug(f"Labels symbols table: {labels_symbols_table}")
    _log.debug(f"Variables symbols table: {variables_symbols_table}")
    return {**labels_symbols_table, **variables_symbols_table}



import logging
from assembler.arguments import args_parser
from assembler.parser import parse
from assembler.code import assemble

_log = logging.getLogger(name=__name__)


def main():
    args = args_parser()

    if args.debug:
        level = getattr(logging, 'DEBUG', None)
    else:
        level = getattr(logging, 'INFO', None)

    logging.basicConfig(level=level)

    parsed_code = parse(args.file)
    _log.debug(f"parsed code: {parsed_code}")

    assembled_code = assemble(parsed_code, args.file)
    _log.debug(f"assembled code: {assembled_code}")


if __name__ == '__main__':
    main()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License

from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='hack-assembler',
    version='0.3.1',
    author="itd27m01",
    author_email="igortiunov@gmail.com",
    description="nand2tetris hack assembler",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ITD27M01/hack-assembler",
    packages=['assembler'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
    entry_points='''
        [console_scripts]
        assembler=assembler.cli:main
    '''
)



import argparse


def args_parser():
    parser = argparse.ArgumentParser(description='Generates hack machine binary code from symbolic form.')
    parser.add_argument('file', type=str, action='store',
                        help='File path with assembly code')

    parser.add_argument('--debug', default=False, action='store_true',
                        help='debug process of assembler')

    return parser.parse_args()
